import contextlib
import datetime
import io
import os
import shutil
import subprocess
import unittest
from unittest.mock import patch

from simple_slurm import Slurm


class Testing(unittest.TestCase):
    script = r"""#!/bin/sh

#SBATCH --array               3-11
#SBATCH --cpus-per-task       15
#SBATCH --dependency          after:65541,afterok:34987
#SBATCH --gres                gpu:kepler:2,gpu:tesla:2,mps:400
#SBATCH --ignore-pbs          
#SBATCH --job-name            name
#SBATCH --output              %A_%a.out
#SBATCH --time                1-02:03:04
"""
    job_file_test_19 = r"""#!/bin/sh

#SBATCH --contiguous          

echo Hello!
"""

    commands = r"""module load python
python run.py 100
echo "done"
"""

    def test_01_args_short(self):
        slurm = Slurm(
            "-a",
            "3-11",
            "-c",
            "15",
            "-J",
            "name",
            "-d",
            "after:65541,afterok:34987",
            "-o",
            r"%A_%a.out",
            "-t",
            "1-02:03:04",
            "--gres",
            "gpu:kepler:2,gpu:tesla:2,mps:400",
            "--ignore_pbs",
            True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_02_args_long(self):
        slurm = Slurm(
            "--array",
            "3-11",
            "--cpus_per_task",
            "15",
            "--job_name",
            "name",
            "--dependency",
            "after:65541,afterok:34987",
            "--output",
            r"%A_%a.out",
            "--time",
            "1-02:03:04",
            "--gres",
            "gpu:kepler:2,gpu:tesla:2,mps:400",
            "--ignore_pbs",
            True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_03_args_simple(self):
        slurm = Slurm(
            "array",
            "3-11",
            "cpus_per_task",
            "15",
            "job_name",
            "name",
            "dependency",
            "after:65541,afterok:34987",
            "output",
            r"%A_%a.out",
            "time",
            "1-02:03:04",
            "gres",
            "gpu:kepler:2,gpu:tesla:2,mps:400",
            "ignore_pbs",
            True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_04_kwargs(self):
        slurm = Slurm(
            array="3-11",
            cpus_per_task="15",
            job_name="name",
            dependency="after:65541,afterok:34987",
            output=r"%A_%a.out",
            time="1-02:03:04",
            gres="gpu:kepler:2,gpu:tesla:2,mps:400",
            ignore_pbs=True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_05_add_arguments_single(self):
        slurm = Slurm()
        slurm.add_arguments(
            array="3-11",
            cpus_per_task="15",
            job_name="name",
            dependency="after:65541,afterok:34987",
            output=r"%A_%a.out",
            time="1-02:03:04",
            gres="gpu:kepler:2,gpu:tesla:2,mps:400",
            ignore_pbs=True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_06_add_arguments_multiple(self):
        slurm = Slurm()
        slurm.add_arguments(array="3-11")
        slurm.add_arguments(cpus_per_task="15")
        slurm.add_arguments(job_name="name")
        slurm.add_arguments(dependency="after:65541,afterok:34987")
        slurm.add_arguments(output=r"%A_%a.out")
        slurm.add_arguments(time="1-02:03:04")
        slurm.add_arguments(gres="gpu:kepler:2,gpu:tesla:2,mps:400")
        slurm.add_arguments(ignore_pbs=True)
        self.assertEqual(self.script, str(slurm))

    def test_07_setter_methods(self):
        slurm = Slurm()
        slurm.set_array("3-11")
        slurm.set_cpus_per_task("15")
        slurm.set_job_name("name")
        slurm.set_dependency("after:65541,afterok:34987")
        slurm.set_output(r"%A_%a.out")
        slurm.set_time("1-02:03:04")
        slurm.set_gres("gpu:kepler:2,gpu:tesla:2,mps:400")
        slurm.set_ignore_pbs(True)
        self.assertEqual(self.script, str(slurm))

    def test_08_parse_range(self):
        slurm = Slurm(
            array=range(3, 12),
            cpus_per_task="15",
            job_name="name",
            dependency="after:65541,afterok:34987",
            output=r"%A_%a.out",
            time="1-02:03:04",
            gres="gpu:kepler:2,gpu:tesla:2,mps:400",
            ignore_pbs=True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_09_parse_int(self):
        slurm = Slurm(
            array=range(3, 12),
            cpus_per_task=15,
            job_name="name",
            dependency="after:65541,afterok:34987",
            output=r"%A_%a.out",
            time="1-02:03:04",
            gres="gpu:kepler:2,gpu:tesla:2,mps:400",
            ignore_pbs=True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_10_parse_dict(self):
        slurm = Slurm(
            array=range(3, 12),
            cpus_per_task=15,
            job_name="name",
            dependency=dict(after=65541, afterok=34987),
            output=r"%A_%a.out",
            time="1-02:03:04",
            gres="gpu:kepler:2,gpu:tesla:2,mps:400",
            ignore_pbs=True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_11_filename_patterns(self):
        slurm = Slurm(
            array=range(3, 12),
            cpus_per_task=15,
            job_name="name",
            dependency=dict(after=65541, afterok=34987),
            output=f"{Slurm.JOB_ARRAY_MASTER_ID}_{Slurm.JOB_ARRAY_ID}.out",
            time="1-02:03:04",
            gres="gpu:kepler:2,gpu:tesla:2,mps:400",
            ignore_pbs=True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_12_output_env_vars_object(self):
        slurm = Slurm()
        self.assertEqual(slurm.SLURM_ARRAY_TASK_ID, r"$SLURM_ARRAY_TASK_ID")

    def test_13_output_env_vars(self):
        self.assertEqual(Slurm.SLURM_ARRAY_TASK_ID, r"$SLURM_ARRAY_TASK_ID")

    def test_14_srun_returncode(self):
        slurm = Slurm(contiguous=True)
        if shutil.which("srun") is not None:
            code = slurm.srun("echo Hello!")
        else:
            with patch("subprocess.run", subprocess_srun):
                code = slurm.srun("echo Hello!")
        self.assertEqual(code, 0)

    def test_15_sbatch_execution(self):
        with io.StringIO() as buffer:
            with contextlib.redirect_stdout(buffer):
                slurm = Slurm(contiguous=True)
                if shutil.which("sbatch") is not None:
                    job_id = slurm.sbatch("echo Hello!")
                else:
                    with patch("subprocess.run", subprocess_sbatch):
                        job_id = slurm.sbatch("echo Hello!")
                stdout = buffer.getvalue()

        out_file = f"slurm-{job_id}.out"
        while True:  # wait for job to finalize
            if os.path.isfile(out_file):
                break
        with open(out_file, "r") as fid:
            contents = fid.read()
        os.remove(out_file)
        self.assertFalse(slurm.is_parsable)
        self.assertIsInstance(job_id, int)
        self.assertIn("Hello!", contents)
        self.assertIn(f"Submitted batch job {job_id}", stdout)

    def test_16_parse_timedelta(self):
        slurm = Slurm(
            array=range(3, 12),
            cpus_per_task=15,
            job_name="name",
            dependency=dict(after=65541, afterok=34987),
            output=r"%A_%a.out",
            time=datetime.timedelta(days=1, hours=2, minutes=3, seconds=4),
            gres="gpu:kepler:2,gpu:tesla:2,mps:400",
            ignore_pbs=True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_17_parse_iterator(self):
        slurm = Slurm(
            array=[range(3, 12)],
            cpus_per_task=15,
            job_name="name",
            dependency=dict(after=65541, afterok=34987),
            output=r"%A_%a.out",
            time=datetime.timedelta(days=1, hours=2, minutes=3, seconds=4),
            gres=("gpu:kepler:2", dict(gpu=dict(tesla=2), mps=400)),
            ignore_pbs=True,
        )
        self.assertEqual(self.script, str(slurm))

    def test_18_false_boolean_arguments(self):
        slurm = Slurm(
            array=[range(3, 12)],
            cpus_per_task=15,
            job_name="name",
            dependency=dict(after=65541, afterok=34987),
            output=r"%A_%a.out",
            time=datetime.timedelta(days=1, hours=2, minutes=3, seconds=4),
            gres=("gpu:kepler:2", dict(gpu=dict(tesla=2), mps=400)),
            ignore_pbs=True,
            wait=False,
        )
        self.assertEqual(self.script, str(slurm))

    def test_19_sbatch_execution_with_job_file(self):
        job_file = "script.sh"
        with io.StringIO() as buffer:
            with contextlib.redirect_stdout(buffer):
                slurm = Slurm(contiguous=True)
                if shutil.which("sbatch") is not None:
                    job_id = slurm.sbatch("echo Hello!", job_file=job_file)
                else:
                    with patch("subprocess.run", subprocess_sbatch):
                        job_id = slurm.sbatch("echo Hello!", job_file=job_file)
                stdout = buffer.getvalue()

        self.assertIsInstance(job_id, int)

        out_file = f"slurm-{job_id}.out"
        while True:  # wait for job to finalize
            if os.path.isfile(out_file):
                break
        # Assert the script was written correctly
        with open(job_file, "r") as fid:
            job_contents = fid.read()
        os.remove(job_file)

        self.assertEqual(job_contents, self.job_file_test_19)

        # Assert the script was executed correctly
        with open(out_file, "r") as fid:
            contents = fid.read()
        os.remove(out_file)

        self.assertIn("Hello!", contents)
        self.assertIn(f"Submitted batch job {job_id}", stdout)

    def test_20_add_cmd_single(self):
        slurm = Slurm(
            "-a",
            "3-11",
            "-c",
            "15",
            "-J",
            "name",
            "-d",
            "after:65541,afterok:34987",
            "-o",
            r"%A_%a.out",
            "-t",
            "1-02:03:04",
            "--gres",
            "gpu:kepler:2,gpu:tesla:2,mps:400",
            "--ignore_pbs",
            True,
        )
        slurm.add_cmd(
            "\n".join(
                (
                    "module load python",
                    "python run.py 100",
                    'echo "done"',
                )
            )
        )
        self.assertEqual(self.script + "\n" + self.commands, str(slurm))

    def test_21_add_cmd_multiple(self):
        slurm = Slurm(
            "-a",
            "3-11",
            "-c",
            "15",
            "-J",
            "name",
            "-d",
            "after:65541,afterok:34987",
            "-o",
            r"%A_%a.out",
            "-t",
            "1-02:03:04",
            "--gres",
            "gpu:kepler:2,gpu:tesla:2,mps:400",
            "--ignore_pbs",
            True,
        )
        slurm.add_cmd("module load python")
        slurm.add_cmd("python run.py 100")
        slurm.add_cmd('echo "done"')
        self.assertEqual(self.script + "\n" + self.commands, str(slurm))

    def test_22_parsable_sbatch_execution(self):
        with io.StringIO() as buffer:
            with contextlib.redirect_stdout(buffer):
                slurm = Slurm(contiguous=True, parsable=True)
                if shutil.which("sbatch") is not None:
                    job_id = slurm.sbatch("echo Hello!")
                else:
                    with patch("subprocess.run", subprocess_sbatch_parsable):
                        job_id = slurm.sbatch("echo Hello!")
                stdout = buffer.getvalue()

        out_file = f"slurm-{job_id}.out"
        while True:  # wait for job to finalize
            if os.path.isfile(out_file):
                break
        with open(out_file, "r") as fid:
            contents = fid.read()
        os.remove(out_file)
        self.assertTrue(slurm.is_parsable)
        self.assertIsInstance(job_id, int)
        self.assertIn("Hello!", contents)
        self.assertEqual(f"{job_id}\n", stdout)


def subprocess_srun(*args, **kwargs):
    print("Hello!!!")
    return subprocess.CompletedProcess(*args, returncode=0)


def subprocess_sbatch(*args, **kwargs):
    job_id = 1234
    out_file = f"slurm-{job_id}.out"
    with open(out_file, "w") as fid:
        fid.write("Hello!!!\n")
    stdout = f"Submitted batch job {job_id}"
    return subprocess.CompletedProcess(
        *args, returncode=1, stdout=stdout.encode("utf-8")
    )


def subprocess_sbatch_parsable(*args, **kwargs):
    job_id = 1234
    out_file = f"slurm-{job_id}.out"
    with open(out_file, "w") as fid:
        fid.write("Hello!!!\n")
    stdout = str(job_id)
    return subprocess.CompletedProcess(
        *args, returncode=0, stdout=stdout.encode("utf-8")
    )


if __name__ == "__main__":
    unittest.main()



from .__about__ import __version__

from .core import Slurm

# create a dummy Slurm object, this forces the creation of attributes for
# file patterns and output environment variables
_ = Slurm()

__all__ = [
    "__version__",
]



import os
import subprocess
import csv
from io import StringIO


class SlurmSqueueWrapper:
    def __init__(self):
        self.command = "squeue"
        self.default_format = '"%i","%j","%t","%M","%L","%D","%C","%m","%b","%R"'
        self.output_format = os.getenv("SQUEUE_FORMAT", self.default_format)

        if not self._is_valid_csv_format(self.output_format):
            raise ValueError("Invalid CSV format in SQUEUE_FORMAT environment variable")

        self.jobs = {}

    def update_squeue(self):
        """Refresh the information from the current queue for the current user"""
        result = subprocess.run(
            [self.command, "--me", "-o", self.output_format],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        if result.returncode != 0:
            raise RuntimeError(f"Error running squeue: {result.stderr}")

        self.jobs = self._parse_output(result.stdout)

    def _is_valid_csv_format(self, format_str: str):
        """validates that the output is a valid csv"""
        try:
            sniffer = csv.Sniffer()
            dialect = sniffer.sniff(format_str, delimiters=",")
            dialect.strict = True
            csv.reader(StringIO(format_str), dialect=dialect)
            return True
        except csv.Error:
            return False

    def _parse_output(self, output: str):
        """converts the stdout into a python dictionary
        each key is a jobid as integer
        """
        csv_file = StringIO(output.strip())
        reader = csv.DictReader(
            csv_file, delimiter=",", quotechar='"', skipinitialspace=True
        )
        jobs = {}
        for row in reader:
            jobs[int(row["JOBID"])] = row
        return jobs

    def display_jobs(self):
        """prints out all job information"""
        for job in self.jobs.values():
            print(job)

    def get_filtered_jobs(self, name_seek: str):
        """Filters jobs by name"""
        matching_jobs = {}
        for job_id, job in self.jobs.items():
            if name_seek in job["NAME"]:
                matching_jobs[job_id] = job
        return matching_jobs



import contextlib
import io
import os
import shutil
import subprocess
import sys
import unittest
from unittest.mock import patch

from simple_slurm.cli import cli


class Testing(unittest.TestCase):
    script = r"""#!/bin/sh

#SBATCH --contiguous          
#SBATCH --job-name            name
#SBATCH --time                00:01:00
""".strip()

    def test_01_using_equal(self):
        self.__run_test(
            [
                "simple_slurm",
                "--job_name=name",
                "--time=00:01:00",
                "--contiguous",
                "echo Hello!!!",
            ]
        )

    def test_02_using_spaces(self):
        self.__run_test(
            [
                "simple_slurm",
                "--job_name",
                "name",
                "--time",
                "00:01:00",
                "--contiguous",
                "echo Hello!!!",
            ]
        )

    def test_03_using_equal_and_spaces(self):
        self.__run_test(
            [
                "simple_slurm",
                "--job_name=name",
                "--time",
                "00:01:00",
                "--contiguous",
                "echo Hello!!!",
            ]
        )

    def __run_test(self, args):
        stdout = run_cli(args)
        script, job_id = parse_stdout(stdout)

        out_file = f"slurm-{job_id}.out"
        while True:  # wait for job to finalize
            if os.path.isfile(out_file):
                break
        with open(out_file, "r") as fid:
            contents = fid.read()
        os.remove(out_file)

        self.assertEqual(self.script, script)
        self.assertIn("Hello!", contents)
        self.assertIn(f"Submitted batch job {job_id}", stdout)


def run_cli(testargs):
    with io.StringIO() as buffer:
        with contextlib.redirect_stdout(buffer):
            with patch.object(sys, "argv", testargs):
                if shutil.which("sbatch") is not None:
                    cli()
                else:
                    with patch("subprocess.run", subprocess_sbatch):
                        cli()
                stdout = buffer.getvalue()
                return stdout


def subprocess_sbatch(*args, **kwargs):
    job_id = 1234
    out_file = f"slurm-{job_id}.out"
    with open(out_file, "w") as fid:
        fid.write("Hello!!!\n")
    stdout = f"Submitted batch job {job_id}"
    return subprocess.CompletedProcess(
        *args, returncode=1, stdout=stdout.encode("utf-8")
    )


def parse_stdout(stdout):
    *script, _, job_msg = stdout.strip().split("\n")
    script = "\n".join(script).strip()
    job_id = int(job_msg.replace("Submitted batch job ", ""))
    return script, job_id


if __name__ == "__main__":
    unittest.main()



import argparse

from .core import Slurm, fmt_key, read_simple_txt


def cli():
    __sbatch = "COMMAND_TO_RUN_WITH_SBATCH"

    # initialize parser
    parser = argparse.ArgumentParser(add_help=False)

    # add arguments into argparser
    arguments = read_simple_txt("arguments.txt")
    arguments_help = read_simple_txt("arguments_help.txt", False)
    for keys, help in zip(arguments, arguments_help):
        parser.add_argument(
            *(fmt_key(k) for k in keys), action=None if help else "store_true"
        )

    # add positional argument for sbatch command
    parser.add_argument(__sbatch, type=str)

    # retrieve given arguments into dict
    kwargs = vars(parser.parse_args())
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    # retrieve sbatch command and remove it from kwargs dict
    cmd = kwargs.pop(__sbatch)

    # populate the Slurm object with the provided arguments,
    # display and execute the slurm script
    slurm = Slurm(**kwargs)
    print(slurm)
    print(cmd)
    slurm.sbatch(cmd)



import subprocess
from datetime import datetime, timedelta
import logging

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class SlurmScancelWrapper:
    sigmtems = {}
    sigmkills = {}
    stale_delta = None

    def __init__(self, staledelta=timedelta(minutes=30)):
        self.stale_delta = staledelta

    def cancel_job(self, job_id: int):
        """Sends a straightforward scancel to a job"""
        job_id = str(job_id)
        result = subprocess.run(
            ["scancel", job_id],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        if result.returncode != 0:
            raise RuntimeError(f"Error cancelling job: {result.stderr.strip()}")

    def signal_job(self, job_id: int):
        """First time it is sent to a job, tries send a SIGTERM to the job id
        If sent again to the same job, attempts a SIGKILL instead
        if that fails as well, involkes scancel without term arguments
        """
        job_id = str(job_id)
        self.prune_old_jobs()
        signal = "--signal=TERM"
        if job_id not in self.sigmtems:
            self.sigmtems[job_id] = datetime.now()
        elif job_id in self.sigmtems:
            signal = "--signal=KILL"
            self.sigmkills[job_id] = datetime.now()
            logger.warning(f"Failed to SIGTERM {job_id}. Sending SIGKILL")
        # Just straight up kills the node via slurm
        elif job_id in self.sigmkills:
            signal = ""
            logger.warning(f"Failed to SIGKILL {job_id}. Terminating with scancel")
        result = subprocess.run(
            ["scancel", signal, job_id],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        if result.returncode != 0:
            raise RuntimeError(f"Error cancelling job: {result.stderr.strip()}")

    def prune_old_jobs(self):
        """Clears out signals information older than self.stale_delta"""
        for signals in [self.sigmtems, self.sigmkills]:
            for job_id in signals:
                if self.sigmtems[job_id] < datetime.now() - self.stale_delta:
                    del self.sigmtems[job_id]

    def cancel_all(self):
        """Cancels all jobs from the current user"""
        result = subprocess.run(
            ["scancel", "--me"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        if result.returncode != 0:
            raise RuntimeError(f"Error cancelling job: {result.stderr.strip()}")



from setuptools import find_packages, setup

with open("README.md", "r") as fid:
    long_description = fid.read()

with open("simple_slurm/__about__.py", "r") as fid:
    __version__ = None
    exec(fid.read())  # loads __version__

setup(
    name="simple_slurm",
    version=__version__,
    author="Arturo Mendoza",
    description="A simple Python wrapper for Slurm with flexibility in mind.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/amq92/simple_slurm",
    packages=find_packages(),
    package_data={"": ["*.txt"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    license="GNU Affero General Public License v3",
    entry_points=dict(console_scripts=["simple_slurm=simple_slurm.cli:cli"]),
)



import argparse
import datetime
import math
import os
import subprocess
from typing import Iterable

from simple_slurm.squeue import SlurmSqueueWrapper
from simple_slurm.scancel import SlurmScancelWrapper

IGNORE_BOOLEAN = "IGNORE_BOOLEAN"


class Slurm:
    """Simple Slurm class for running sbatch commands.

    See https://slurm.schedmd.com/sbatch.html for a complete list of arguments
    accepted by the sbatch command (ex. -a, --array).

    Validation of arguments is handled by the argparse module.

    Multiple syntaxes are allowed for defining the arguments.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the parser with the given arguments."""

        # initialize parser
        self.namespace = Namespace()
        self.parser = argparse.ArgumentParser()
        self.squeue = SlurmSqueueWrapper()
        self.scancel = SlurmScancelWrapper()

        # add arguments into argparser
        for keys in read_simple_txt("arguments.txt"):
            self.parser.add_argument(*(fmt_key(k) for k in keys))

        # create setter methods for each argument
        for keys in read_simple_txt("arguments.txt"):
            create_setter_method(keys[0])

        # add filename patterns as static variables
        for pattern in read_simple_txt("filename_patterns.txt"):
            setattr(Slurm, *pattern)

        # add output environment variables as static variables
        for (var,) in read_simple_txt("output_env_vars.txt"):
            setattr(Slurm, var, "$" + var)

        # add provided arguments in constructor
        self.add_arguments(*args, **kwargs)

        # contain a list of "single-line" commands to dispatch
        self.run_cmds = []

    def __str__(self) -> str:
        """Print the generated sbatch script."""
        return self.script()

    def __repr__(self) -> str:
        """Print the argparse namespace."""
        params = dict(vars(self.namespace))  # make copy
        params["run_cmds"] = self.run_cmds
        return repr(params)

    def _add_one_argument(self, key: str, value: str):
        """Parse the given key-value pair (the argument is given in key)."""
        key, value = fmt_key(key), fmt_value(value)
        if value is not IGNORE_BOOLEAN:
            self.parser.parse_args([key, value], namespace=self.namespace)

    def add_arguments(self, *args, **kwargs):
        """Parse the given key-value pairs.

        Both syntaxes *args and **kwargs are allowed, ex:
            add_arguments('arg1', val1, 'arg2', val2, arg3=val3, arg4=val4)
        """
        for key, value in zip(args[0::2], args[1::2]):
            self._add_one_argument(key, value)
        for key, value in kwargs.items():
            self._add_one_argument(key, value)
        return self

    def add_cmd(self, *cmd_args: str):
        """Add a new command to the command list, it can be provided as a single
        argument (ie. a string) or a collection of arguments (all converted to
        strings and spaces are added in-between).

        For example, these syntaxes are equivalent
            > slurm.add_cmd('python main.py --input 1')
            > slurm.add_cmd('python', 'main.py', '--input', 1)
        """
        cmd = " ".join([str(cmd) for cmd in cmd_args]).strip()
        if len(cmd):
            self.run_cmds.append(cmd)
        return self

    def reset_cmd(self):
        """Reset the command list"""
        self.run_cmds = []

    @staticmethod
    def _valid_key(key: str) -> str:
        """Long arguments (for slurm) constructed with '-' have been internally
        represented with '_' (for Python). Correct for this in the output.
        """
        return key.replace("_", "-")

    def script(self, shell: str = "/bin/sh", convert: bool = True):
        """Generate the sbatch script for the current arguments and commands"""

        arguments = (
            "\n".join(
                (
                    f"#!{shell}",
                    "",
                    *(
                        f"#SBATCH --{self._valid_key(k):<19} {v}"
                        for k, v in vars(self.namespace).items()
                        if v is not None
                    ),
                )
            )
            + "\n"
        )
        commands = "\n".join(
            [cmd.replace("$", "\\$") if convert else cmd for cmd in self.run_cmds]
        )
        script = "\n".join((arguments, commands)).strip() + "\n"
        return script

    def srun(self, run_cmd: str, srun_cmd: str = "srun") -> int:
        """Run the srun command with all the (previously) set arguments and
        the provided command in 'run_cmd'.
        """
        args = (
            f"--{self._valid_key(k)}" + f"={v}" if len(v) else ""
            for k, v in vars(self.namespace).items()
            if v is not None
        )
        cmd = " ".join((srun_cmd, *args, run_cmd))

        result = subprocess.run(cmd, shell=True, check=True)
        return result.returncode

    @property
    def is_parsable(self) -> bool:
        return getattr(self.namespace, "parsable", None) is not None

    def sbatch(
        self,
        *run_cmd: str,
        convert: bool = True,
        verbose: bool = True,
        sbatch_cmd: str = "sbatch",
        shell: str = "/bin/sh",
        job_file: str = None,
    ) -> int:
        """Run the sbatch command with all the (previously) set arguments and
        the provided command in 'run_cmd' alongside with the previously set
        commands using 'add_cmd'.

        Note that 'run_cmd' can accept multiple arguments. Thus, any of the
        other arguments must be given as key-value pairs :
            > slurm.sbatch('python main.py')
            > slurm.sbatch('python', 'main.py')
            > slurm.sbatch('python', 'main.py', verbose=False)
        This function employs the 'here document' syntax, which requires that
        bash variables be scaped. This behavior is default, set 'convert'
        to False to disable it.

        This function employs the following syntax:
            $ slurm_cmd << EOF
            > bash_script
            > run_command
            > EOF

        For such reason if any bash variable is employed by the 'run_command',
        the '$' should be scaped into '\$'. This behavior is default, set
        'convert' to False to disable it.

        If the argument 'job_file' is used, the script will be written to the
        designated file, and then the command `sbatch <job_file>` will be
        executed.
        """
        self.add_cmd(*run_cmd)
        if job_file is not None:
            with open(job_file, "w") as fid:
                fid.write(self.script(shell, convert))
            cmd = sbatch_cmd + " " + job_file
        else:
            cmd = "\n".join(
                (
                    sbatch_cmd + " << EOF",
                    self.script(shell, convert),
                    "EOF",
                )
            )
        result = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE)
        # init for clarity
        job_id = None
        stdout = ""
        if self.is_parsable:
            # gather the first line from stdout
            stdout = result.stdout.decode().strip()
            # parsable will be of format job_id[:cluster]
            # ref: https://slurm.schedmd.com/sbatch.html#OPT_parsable
            job_id = int(stdout.split(":")[0])
            assert result.returncode == 0, result.stderr
        else:
            success_msg = "Submitted batch job"
            stdout = result.stdout.decode()
            assert success_msg in stdout, result.stderr
            job_id = int(stdout.split(" ")[3])
        assert job_id is not None, "this should never happen, assert for linter"
        if verbose:
            print(stdout)
        return job_id


class Namespace:
    """Dummy class required for accessing the arguments in argparse"""

    pass


def create_setter_method(key: str):
    """Creates the setter method for the given 'key' attribute of a Slurm
    object
    """

    def set_key(self, value):
        return self.add_arguments(key, value)

    set_key.__name__ = f"set_{key}"
    set_key.__doc__ = f'Setter method for the argument "{key}"'
    setattr(Slurm, set_key.__name__, set_key)


def fmt_key(key: str) -> str:
    """Maintain correct formatting for keys in key-value pairs"""
    key = str(key).strip()
    if "-" not in key:
        key = f"--{key}" if len(key) > 1 else f"-{key}"
    return key


def fmt_value(value) -> str:
    """Maintain correct formatting for values in key-value pairs
    This function handles some special cases for the type of value:
        1) A 'range' object:
            Converts range(3, 15) into '3-14'.
            Useful for defining job arrays using a Python syntax.
            Note the correct form of handling the last element.
        2) A 'dict' object:
            Converts dict(after=65541, afterok=34987)
            into 'after:65541,afterok:34987'.
            Useful for arguments that have multiple 'sub-arguments',
            such as when declaring dependencies.
        3) A `datetime.timedelta` object:
            Converts timedelta(days=1, hours=2, minutes=3, seconds=4)
            into '1-02:03:04'.
            Useful for arguments involving time durations.
        4) An `iterable` object:
            Will recursively format each item
            Useful for defining lists of parameters
    """
    if isinstance(value, str):
        pass

    elif isinstance(value, range):
        start, stop, step = value.start, value.stop - 1, value.step
        value = f"{start}-{stop}" + ("" if value.step == 1 else f":{step}")

    elif isinstance(value, dict):
        value = ",".join((f"{k}:{fmt_value(v)}" for k, v in value.items()))

    elif isinstance(value, datetime.timedelta):
        time_format = "{days}-{hours2}:{minutes2}:{seconds2}"
        value = format_timedelta(value, time_format=time_format)

    elif isinstance(value, Iterable):
        value = ",".join((fmt_value(item) for item in value))

    elif isinstance(value, bool):
        value = "" if value else IGNORE_BOOLEAN

    return str(value).strip()


def read_simple_txt(path: str, split: bool = True) -> list:
    """Simple function for reading the txt files."""
    __pkg_path = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(__pkg_path, path), "r") as fid:
        lines = fid.read().split("\n")
        if not split:
            return lines
        return [[wrd.strip() for wrd in ln.split(",")] for ln in lines]


def format_timedelta(value: datetime.timedelta, time_format: str):
    """Format a datetime.timedelta (https://stackoverflow.com/a/30339105)"""
    if hasattr(value, "seconds"):
        seconds = value.seconds + value.days * 24 * 3600
    else:
        seconds = int(value)

    seconds_total = seconds

    minutes = int(math.floor(seconds / 60))
    minutes_total = minutes
    seconds -= minutes * 60

    hours = int(math.floor(minutes / 60))
    hours_total = hours
    minutes -= hours * 60

    days = int(math.floor(hours / 24))
    days_total = days
    hours -= days * 24

    years = int(math.floor(days / 365))
    years_total = years
    days -= years * 365

    return time_format.format(
        **{
            "seconds": seconds,
            "seconds2": str(seconds).zfill(2),
            "minutes": minutes,
            "minutes2": str(minutes).zfill(2),
            "hours": hours,
            "hours2": str(hours).zfill(2),
            "days": days,
            "years": years,
            "seconds_total": seconds_total,
            "minutes_total": minutes_total,
            "hours_total": hours_total,
            "days_total": days_total,
            "years_total": years_total,
        }
    )
